from dalloriam import docker

from tests.mocks.location import mock_location

from typing import cast

from unittest import mock


@mock.patch.object(docker.client.DockerClient, '_login')
def test_client_init_calls_login_when_required(mock_login: mock.MagicMock):

    docker.Client()

    mock_login.assert_not_called()

    docker.Client(server='some/custom/server')

    mock_login.assert_called_once()
    mock_login.reset_mock()

    docker.Client(username='hello')
    mock_login.assert_not_called()

    docker.Client(username='hello', password='world')
    mock_login.assert_called_once()


def test_client_build_calls_shell_run_properly():
    client = docker.Client()

    with mock.patch.object(docker.client.system, 'location', mock_location),\
             mock.patch('sh.docker', create=True) as mock_run:
        client.build('some_path', 'my_image', 'mytag')
        mock_run.assert_called_once()

        mock_run.assert_called_with('build', '-t', 'my_image:mytag', '.')


def test_client_build_sets_default_tag():
    client = docker.Client()

    with mock.patch.object(docker.client.system, 'location', mock_location), \
            mock.patch('sh.docker', create=True) as mock_run:
        client.build('some_path', 'my_image')
        mock_run.assert_called_once()

        mock_run.assert_called_with('build', '-t', 'my_image:latest', '.')


def test_client_push_calls_correct_command():
    client = docker.Client()

    with mock.patch('sh.docker', create=True) as mock_run:
        client.push('someuser/some_image', 'some_tag')
        mock_run.assert_called_once()

        mock_run.assert_called_with('push', 'someuser/some_image:some_tag')


def test_client_container_initializes_container_correctly():
    client = docker.Client()

    port_data = {32: 32}
    vol_data = {'vol': 'vol'}

    with mock.patch.object(docker.client.Container, 'start') as mock_start, \
            mock.patch.object(docker.client.Container, 'stop') as mock_stop:

        with client.container('some_image', 'some_tag', ports=port_data, volumes=vol_data) as c:
            assert isinstance(c, docker.Container)
            assert c.name.startswith('some_image') and c.name != 'some_image'
            mock_start.assert_called_once_with(port_data, vol_data)
            mock_stop.assert_not_called()

        mock_stop.assert_called_once()


def test_client_login_sends_correct_request():
    client = docker.Client()

    with mock.patch('sh.docker') as mock_docker:
        client._login()
        mock_docker.assert_called_once_with('login')
        mock_docker.reset_mock()

        client._username = 'hello'
        client._login()
        mock_docker.assert_called_once_with('login', '-u', 'hello')
        mock_docker.reset_mock()

        client._password = 'world'
        client._login()
        mock_docker.assert_called_once_with('login', '-u', 'hello', '-p', 'world')
        mock_docker.reset_mock()

        client._server = 'http://mock.server.ca/cr'
        client._login()
        mock_docker.assert_called_once_with('login', '-u', 'hello', '-p', 'world', 'http://mock.server.ca/cr')



from dalloriam.exceptions import AccessDenied

from dataclasses import dataclass, field

from google.cloud import datastore

from typing import Any, Dict, List

import google.auth.transport.requests
import google.oauth2.id_token


@dataclass
class User:

    _ds = datastore.Client()

    uid: str
    permissions: List[str] = field(default_factory=lambda: [])
    services: Dict[str, Dict[str, Any]] = field(default_factory=lambda: {})

    def save(self) -> None:
        entity = datastore.Entity(User._ds.key('User', self.uid))
        entity.update(**{'permissions': self.permissions, 'services': self.services})
        User._ds.put(entity)

    @property
    def serialized(self) -> Dict[str, Any]:
        """
        Serializes the user info.
        Returns:
            The serialized user.
        """
        return {
            'id': self.uid,
            'permissions': self.permissions,
            'services': self.services
        }

    @staticmethod
    def _try_get_user_info(user_id: str) -> Dict[str, Any]:
        return User._ds.get(User._ds.key('User', user_id))

    @staticmethod
    def _create_default_user(user_id: str) -> Dict[str, Any]:
        entity = datastore.Entity(User._ds.key('User', user_id))
        default_info = {
            'permissions': [],
            'services': {}
        }
        entity.update(**default_info)
        User._ds.put(entity)
        return default_info

    @staticmethod
    def from_uid(user_id: str) -> 'User':
        """
        Fetches the permissions and services from datastore, creating the user over there if required.
        Args:
            user_id (str): The ID of the user from firebase auth.

        Returns:
            The user object.
        """
        user_info = User._try_get_user_info(user_id)

        if user_info is None:
            # Create the user with default
            user_info = User._create_default_user(user_id)

        return User(uid=user_id, **user_info)

    @staticmethod
    def from_token(token: str, refresh_token: str = None) -> 'User':
        """
        Validate an auth token & return the claims. Can only be called in an HTTP context.
        Args:
            token (str): The token to validate.

        Returns:
            The user claims.
        """
        if not token:
            raise AccessDenied('Invalid Token')

        HTTP_REQUEST = google.auth.transport.requests.Request()
        try:
            claims = google.oauth2.id_token.verify_firebase_token(token, HTTP_REQUEST)
        except ValueError:
            raise AccessDenied("Token Expired")  # TODO: Add refresh

        if not claims:
            raise AccessDenied("Invalid Token")

        return User.from_uid(claims['user_id'])



import subprocess
import sys


def notification(title: str, message: str) -> None:
    """
    Sends a system notification.
    Args:
        title (str): The title of the notification
        message (str): The notification body.
    """
    if sys.platform == 'linux':
        subprocess.Popen(['notify-send', title, message])
    else:
        # TODO: Support macOS
        raise NotImplementedError



from dalloriam import docker

from unittest import mock


def test_container_splits_name_properly():
    with mock.patch.object(docker.container, '_random_name', return_value='asdf'):
        c = docker.Container('some.domain.io/some_directory/some_image', 'latest')
        assert c.name == 'some_image-asdf'



from abc import ABCMeta, abstractmethod

from copy import deepcopy

from itertools import product

from numbers import Real

from typing import Any, Dict, Iterator, List


class Variation(metaclass=ABCMeta):

    @abstractmethod
    def expand(self) -> List[Any]:
        """ Expands the variation to a list of values. """


class RangeVariation(Variation):

    def __init__(self, min_val: Real, max_val: Real, step: Real) -> None:
        self._min = min_val
        self._max = max_val
        self._step = step

    def expand(self) -> List[Any]:
        """ Expands the variation to a list of values. """
        values = []
        i = self._min
        while i < self._max:
            values.append(i)
            i += self._step
        values.append(self._max)

        return values


class ValuesVariation(Variation):

    def __init__(self, values: List[Any]) -> None:
        self._values = values

    def expand(self) -> List[Any]:
        return self._values


def explode_variations(base_data: Dict[str, Any], variations: Dict[str, Variation]) -> Iterator[Dict[Any, Any]]:
    """ Blows up variations into a list from a range dict and some default parameters. """
    keys = []  # Preserve order
    value_ranges = []
    for k, var_object in variations.items():
        keys.append(k)
        value_ranges.append(var_object.expand())

    combined = list(product(*value_ranges))
    for combination in combined:
        copied = deepcopy(base_data)
        for i, v in enumerate(combination):
            k = keys[i]
            subsect = copied
            while '.' in k:
                prefix, rest = k.split('.', 1)
                k = rest
                subsect = subsect[prefix]
            subsect[k] = v
        yield copied



from typing import Iterable, Dict

import random
import sh
import string


def _random_name(size: int = 12) -> str:
    return ''.join(random.choices(string.ascii_uppercase + string.digits, k=size))


class Container:

    def __init__(self, image_name: str, tag: str) -> None:
        self.name = f'{image_name.split("/")[-1]}-{_random_name()}'
        self._image = f'{image_name}:{tag}'

    def start(self, ports: Dict[int, int] = None, volumes: Dict[str, str] = None) -> None:
        """
        Starts the container.
        Args:
            ports (Dict[int, int]): The port mappings.
            volumes (Dict[str, str]): The volume mappings.
        """
        ports_lst = []
        if ports is not None:
            for k, v in ports.items():
                ports_lst.append('-p')
                ports_lst.append(f'{k}:{v}')

        vol_lst = []
        if volumes is not None:
            for k, v in volumes.items():
                vol_lst.append('-v')
                vol_lst.append(f'{k}:{v}')

        sh.docker.run('--rm', '-d', *ports_lst, *vol_lst, '--name', self.name, self._image)

    def stop(self) -> None:
        """
        Stops the container.
        """
        sh.docker.stop(self.name)

    def exec(self, cmd: Iterable[str], background: bool = False) -> None:
        """
        Executes a command inside the container
        Args:
            cmd (Iterable[str]): The command arguments.
            background (bool): Whether to execute the command in the background or to await the results.
        """
        out = sh.docker.exec('-d' if background else '-i', self.name, *cmd)
        if not background:
            print(out)
